// Mitsuba includes
#include <mitsuba/core/fresolver.h>
#include <mitsuba/render/bsdf.h>
#include <mitsuba/render/texture.h>
#include <mitsuba/core/warp.h>

#include <mitsuba/hw/basicshader.h>

// Local includes
#include "microfacet.h"
#include "ior.h"
#include "layered.hpp"

MTS_NAMESPACE_BEGIN

class LayeredAnisotropic : public BSDF
{
public:
    LayeredAnisotropic(const Properties &props) : BSDF(props), rng(new Random)
    {
        parseLayers(props, nb_layers, m_etas, m_kappas, m_alphaUs, m_alphaVs);
        MicrofacetDistribution distr(props);
        m_type = distr.getType();
        max_bounces = props.getInteger("max_bounces", 100);
    }

    LayeredAnisotropic(Stream *stream, InstanceManager *manager)
        : BSDF(stream, manager)
    {

        configure();
    }

    void configure()
    {
        unsigned int extraFlags = 0;

        m_components.clear();
        for (int i = 0; i < nb_layers; i++)
        {

            if (m_kappas[i + 1].isZero())
            { // dielectric
                if (m_alphaUs[i] != m_alphaVs[i])
                    extraFlags |= EAnisotropic;

                m_components.push_back(EGlossyReflection | EFrontSide | EBackSide | EUsesSampler | extraFlags);

                m_components.push_back(EGlossyTransmission | EFrontSide | EBackSide | EUsesSampler | ENonSymmetric | extraFlags);

                m_components.push_back(EGlossyReflection | EFrontSide);
            }
            else
            { // conductor
                if (m_alphaUs[i] != m_alphaVs[i])
                    extraFlags |= EAnisotropic;

                m_components.push_back(EGlossyReflection | EFrontSide | extraFlags);
            }

            extraFlags = 0;
        }

        BSDF::configure();
    }

    /* 
    *   Sample a Rough Dielectric layer
    */
    Spectrum sampleRoughDielectric(const Vector3 &wi,
                                   const Spectrum &eta,
                                   const Spectrum &kappa,
                                   float alphaU,
                                   float alphaV,
                                   const Point3 &sample,
                                   Vector3 &wo,
                                   bool &is_reflected) const
    {

        MicrofacetDistribution distr(m_type, alphaU, alphaV, true);

        // Sample a normal
        Vector wh = distr.sample(math::signum(Frame::cosTheta(wi)) * wi, Point2(sample.x, sample.y));

        // Transmitted angle
        Float cosThetaT;

        Float F = fresnelDielectricExt(dot(wi, wh), cosThetaT, eta.average());

        Float weight = 1.0f;

        bool sampleReflection = true;

        if (sample.z > F)
            sampleReflection = false;

        if (sampleReflection)
        { // sample reflection
            is_reflected = true;

            wo = reflect(wi, wh);

            /* Side check */
            if (Frame::cosTheta(wi) * Frame::cosTheta(wo) <= 0)
                return Spectrum(0.0f);
        }
        else
        {
            is_reflected = false;

            if (cosThetaT == 0)
                return Spectrum(0.0f);

            wo = refract(wi, wh, eta.average(), cosThetaT);

            /* Side check */
            if (Frame::cosTheta(wi) * Frame::cosTheta(wo) >= 0)
                return Spectrum(0.0f);

            /* Radiance must be scaled to account for the solid angle compression
               that occurs when crossing the interface. */
            Float factor = cosThetaT < 0 ? 1.0 / eta.average() : eta.average();

            weight *= (factor * factor);
        }

        // Evaluate Shadowing Masking through SmithG1 (since only visible normals were queried)
        Float G = distr.smithG1(wo, wh);

        return Spectrum(weight * G);
    }

    /* 
    *   Sample a rough Conductor layer
    */
    Spectrum sampleRoughCondutor(const Vector3 &wi,
                                 const Spectrum &eta,
                                 const Spectrum &kappa,
                                 Float alphaU,
                                 Float alphaV,
                                 const Point2 &sample,
                                 Vector3 &wo) const
    {

        if (Frame::cosTheta(wi) <= 0)
            return Spectrum(0.0f);

        MicrofacetDistribution distr(m_type, alphaU, alphaV, true);

        Vector wh = distr.sample(wi, sample);

        wo = reflect(wi, wh);

        if (Frame::cosTheta(wo) <= 0)
            return Spectrum(0.0f);

        Spectrum F = fresnelConductorExact(dot(wi, wh), eta, kappa);

        // Evaluate Shadowing Masking through SmithG1 (since only visible normals were queried)
        Float G = distr.smithG1(wo, wh);

        return F * G;
    }

    /* 
    *   Eval a Rough Dielectric layer
    */
    Spectrum evalRoughDielectric(const Vector3 &wi,
                                 const Vector3 &wo,
                                 const Spectrum &eta,
                                 Float alphaU,
                                 Float alphaV) const
    {
        if (Frame::cosTheta(wi) == 0)
            return Spectrum(0.0f);
        /* Determine the type of interaction */
        bool reflection = Frame::cosTheta(wi) * Frame::cosTheta(wo) > 0;

        Vector wh;
        if (reflection)
        {
            /* Calculate the reflection half-vector */
            wh = normalize(wo + wi);
        }
        else
        {
            /* Calculate the transmission half-vector */
            Float _eta = Frame::cosTheta(wi) > 0 ? eta.average() : 1.0 / eta.average();

            wh = normalize(wi + wo * _eta);
        }

        /* Ensure that the half-vector points into the
           same hemisphere as the macrosurface normal */
        wh *= math::signum(Frame::cosTheta(wh));

        /* Construct the microfacet distribution matching the
           roughness values at the current surface position. */
        MicrofacetDistribution distr(m_type, alphaU, alphaV, true);

        /* Evaluate the microfacet normal distribution */
        const Float D = distr.eval(wh);

        /* Fresnel factor */
        const Float F = fresnelDielectricExt(dot(wi, wh), eta.average());

        /* Smith's shadow-masking function */
        const Float G = distr.G(wi, wo, wh);

        if (reflection)
        {
            /* Calculate the total amount of reflection */
            return Spectrum(F * D * G / (4.0f * std::abs(Frame::cosTheta(wi))));
        }
        else
        {
            Float _eta = Frame::cosTheta(wi) > 0.0f ? eta.average() : 1.0 / eta.average();

            /* Calculate the total amount of transmission */
            Float sqrtDenom = dot(wi, wh) + _eta * dot(wo, wh);
            Float value = ((1 - F) * D * G * _eta * _eta * dot(wi, wh) * dot(wo, wh)) /
                          (Frame::cosTheta(wi) * sqrtDenom * sqrtDenom);

            /* Missing term in the original paper: account for the solid angle
               compression when tracing radiance -- this is necessary for
               bidirectional methods */
            Float factor = Frame::cosTheta(wi) > 0 ? 1.0 / eta.average() : eta.average();

            return Spectrum(std::abs(value * factor * factor));
        }
    }

    /* 
    *   Eval a Rough Conductor layer
    */
    Spectrum evalRoughConductor(const Vector3 &wi,
                                const Vector3 &wo,
                                const Spectrum &eta,
                                const Spectrum &kappa,
                                Float alphaU,
                                Float alphaV) const
    {

        if (Frame::cosTheta(wi) <= 0 || Frame::cosTheta(wo) <= 0)
            return Spectrum(0.0f);

        MicrofacetDistribution distr(m_type, alphaU, alphaV, true);

        // Compute Half-way vector
        Vector wh = normalize(wo + wi);

        // Evaluate Microface Distribution
        Float D = distr.eval(wh);

        // Evaluate complex Fresnel
        Spectrum F = fresnelConductorExact(dot(wi, wh), eta, kappa);

        // Evaluate Shadowing Masking
        Float G = distr.G(wi, wo, wh);

        return F * G * D / (4.0f * Frame::cosTheta(wi));
    }

    Float pdfRoughConductor(const Vector3 &wi,
                            const Vector3 &wo,
                            Float alphaU,
                            Float alphaV) const
    {
        if (Frame::cosTheta(wi) <= 0 || Frame::cosTheta(wo) <= 0)
            return 0.0f;

        MicrofacetDistribution distr(m_type, alphaU, alphaV, true);

        // Compute Half-way vector
        Vector wh = normalize(wo + wi);

        return distr.eval(wh) * distr.smithG1(wi, wh) / (4.0f * Frame::cosTheta(wi));
    }

    Float pdfRoughDielectric(const Vector3 &wi,
                             const Vector3 &wo,
                             const Spectrum &eta,
                             Float alphaU,
                             Float alphaV) const
    {

        bool reflection = Frame::cosTheta(wi) * Frame::cosTheta(wo) > 0.0f;

        // Half-way vector
        Vector wh;

        // Jacobian
        Float Jh;

        if (reflection)
        {
            wh = normalize(wi + wo);

            Jh = 1.0f / (4.0f * dot(wi, wh));
        }
        else
        {
            Float _eta = Frame::cosTheta(wi) > 0 ? eta.average() : 1.0 / eta.average();

            wh = normalize(wi + _eta * wo);

            Float sqrtDenom = dot(wi, wh) + _eta * dot(wo, wh);
            Jh = (_eta * _eta * dot(wo, wh)) / (sqrtDenom * sqrtDenom);
        }

        wh *= math::signum(Frame::cosTheta(wh));

        MicrofacetDistribution distr(m_type, alphaU, alphaV, true);

        Float pdf = distr.pdf(math::signum(Frame::cosTheta(wi)) * wi, wh);

        Float F = fresnelDielectricExt(dot(wi, wh), eta.average());

        if (reflection)
            return std::abs(F * pdf * Jh);
        else
            return std::abs((1.0f - F) * pdf * Jh);
    }

    /* We use an additional random generator since bRec.sampler is not offered in eval normally */
    ref<Random> rng;
    float generateRandomNumber() const
    {
        return const_cast<Random *>(rng.get())->nextFloat();
    }

    /*
    *   Evaluates a multi-layered bsdf 
    *   TODOs: - Test with more than 2 layers 
    *          - Add partciping media layer
    *          - Adjust for dielectrics only ( Frame::cosTheta(bRec.wi) <= 0 || Frame::cosTheta(bRec.wo) <= 0)
    */
    Spectrum eval(const BSDFSamplingRecord &bRec, EMeasure measure) const
    {
        if (measure != ESolidAngle || Frame::cosTheta(bRec.wi) <= 0 || Frame::cosTheta(bRec.wo) <= 0)
            return Spectrum(0.0f);

        /* Characteristics of the photon: direction, energy, position in the layered structure. */
        Vector3 wi = bRec.wi;
        int index = 0;
        int bounces = 0;
        Spectrum e = Spectrum(1.0f); // Energy throughput
        Spectrum f = Spectrum(0.0f); // bsdfValue * cosTheta = eval
        bool is_going_up = false;    // boolean indicating if ray is going deeper in the layers or not

        /* Start the iteration loop */
        while (bounces < max_bounces || max_bounces == -1)
        {
            /* Obtain the material information */
            Float alphaU = m_alphaUs[index];
            Float alphaV = m_alphaVs[index];
            Spectrum extEta = m_etas[index];
            Spectrum intEta = m_etas[index + 1];

            Spectrum eta = intEta / extEta;
            Spectrum kappa = m_kappas[index + 1] / extEta;

            /* Next Event Estimation */
            if (index == 0)
            {
                if (kappa.isZero())
                { // Dielectric
                    f += e * this->evalRoughDielectric(wi, bRec.wo, eta, alphaU, alphaV);
                }
                else
                { // Conductor
                    f += e * this->evalRoughConductor(wi, bRec.wo, eta, kappa, alphaU, alphaV);
                }
            }

            /* Get the next interaction with the layer */
            Vector wo_sampled;
            if (kappa.isZero())
            {
                Point3 _sample(generateRandomNumber(), generateRandomNumber(), generateRandomNumber());
                bool is_reflected = true;

                // the sampling takes care of inverting eta given cosTheta(wi)'s sign
                e *= this->sampleRoughDielectric(wi, eta, kappa, alphaU, alphaV, _sample, wo_sampled, is_reflected);

                if (is_reflected)
                {
                    is_going_up = !is_going_up;
                }
            }
            else
            {
                Point2 _sample(generateRandomNumber(), generateRandomNumber());
                e *= this->sampleRoughCondutor(wi, eta, kappa, alphaU, alphaV, _sample, wo_sampled);
                is_going_up = !is_going_up;
            }

            if (e.isZero())
            {
                return Spectrum(0.0f);
            }

            /* If the ray continues its travel, updates its parameters and revert it for the next layer interaction*/
            wi = -wo_sampled;
            bounces += 1;

            /* Update transport information */
            index += (is_going_up) ? -1 : 1;

            // Check if the photon is leaving the layer. This is the only
            // 'winning' situation where we have to record an outgoing direction.
            if (index < 0 || index >= nb_layers || e.isZero())
            {
                Assert(f.isValid());
                return f;
            }
        }
        return f;
    }

    /*
    * Returns the pdf of the first layer as an approximation 
    */
    Float pdf(const BSDFSamplingRecord &bRec, EMeasure measure) const
    {
        // The first layer properties are at index 1 for eta and kappa
        // and index 0 for the roughness
        float alphaU = m_alphaUs[0];
        float alphaV = m_alphaVs[0];
        Spectrum eta1 = m_etas[0]; // external IOR usually air
        Spectrum eta2 = m_etas[1];
        Spectrum kappa2 = m_kappas[1];

        Spectrum eta = eta2 / eta1;
        Spectrum kappa = kappa2 / eta1;

        /* A small epsilon is added to the pdf*/
        if (kappa.isZero())
        { // Dielectric
            return this->pdfRoughDielectric(bRec.wi, bRec.wo, eta, alphaU, alphaV) + Epsilon;
        }
        else
        { // Conductor
            return this->pdfRoughConductor(bRec.wi, bRec.wo, alphaU, alphaV) + Epsilon;
        }
    }

    Spectrum sample(BSDFSamplingRecord &bRec, const Point2 &sample) const
    {
        Float pdf;
        return this->sample(bRec, pdf, sample);
    }

    Spectrum sample(BSDFSamplingRecord &bRec, Float &pdf, const Point2 &sample) const
    {
        if (Frame::cosTheta(bRec.wi) < 0 ||
            ((bRec.component != -1 && bRec.component != 0) ||
             !(bRec.typeMask & EGlossyReflection)))
            return Spectrum(0.0f);

        /* Characteristics of the photon: direction, energy, position in the layered
         * structure. */
        Vector wi = bRec.wi;
        int index = 0;
        int bounces = 0;
        Spectrum e = Spectrum(1.0);
        bool is_going_up = false;

        /* Start the iteration loop */
        while (bounces < max_bounces)
        {

            // Get info about layer
            float alphaU = m_alphaUs[index];
            float alphaV = m_alphaVs[index];
            Spectrum extEta = m_etas[index];
            Spectrum intEta = m_etas[index + 1];

            Spectrum eta = intEta / extEta;
            Spectrum kappa = m_kappas[index + 1] / extEta;

            /* Get the next interaction with the layer */
            Vector3 wo;
            if (kappa.isZero())
            { // Dielectric
                bool is_reflected = true;
                Point3 _sample(bRec.sampler->next1D(), bRec.sampler->next1D(), bRec.sampler->next1D());

                // the sampling takes care of inverting eta given cosTheta(wi)'s sign
                e *= sampleRoughDielectric(wi, eta, kappa, alphaU, alphaV, _sample, wo, is_reflected);
                if (is_reflected)
                {
                    is_going_up = !is_going_up;
                }
            }
            else
            { // Conductor
                Point2 _sample(bRec.sampler->next1D(), bRec.sampler->next1D());
                e *= sampleRoughCondutor(wi, eta, kappa, alphaU, alphaV, _sample, wo);

                is_going_up = !is_going_up;
            }

            if (e.isZero())
            {
                return Spectrum(0.0f);
            }

            /* If the ray continues its travel, updates its parameters and revert it for the next layer interaction*/
            wi = -wo;

            bounces += 1;

            /* Update transport information */
            index += (is_going_up) ? -1 : 1;

            // Check if the photon is leaving the layer. This is the only
            // 'winning' situation where we have to record an outgoing direction.
            if (index < 0)
            {
                // if we are leaving the layer invert the outgoing direction to counter the
                // after layer interaction
                bRec.wo = -wi;
                bRec.eta = 1.0f;
                bRec.sampledComponent = 0;
                bRec.sampledType = EGlossyReflection;
                pdf = this->pdf(bRec, ESolidAngle);
                if (pdf == 0.0f)
                { // pdf cannot be null
                    pdf = 1.0f;
                }

                Assert(e.isValid());
                return e;
            }
            else if (index >= nb_layers)
            {
                return Spectrum(0.0f);
            }
        }

        return Spectrum(0.0f);
    }

    void addChild(const std::string &name, ConfigurableObject *child)
    {

        BSDF::addChild(name, child);
    }

    void serialize(Stream *stream, InstanceManager *manager) const
    {
        // Diffuse template
        BSDF::serialize(stream, manager);
    }

    std::string formStringForLayers(int nb_layers,
                                    std::vector<Spectrum> m_etas,
                                    std::vector<Spectrum> m_kappas,
                                    std::vector<Float> m_alphaUs,
                                    std::vector<Float> m_alphaVs) const
    {

        std::stringstream ss;
        ss << "    "
           << "Microfacet Distribution: " << MicrofacetDistribution::distributionName(m_type) << "\n\n"
           << "    "
           << "Max bounces (-1 = infinity): " << max_bounces << "\n\n"
           << "    "
           << "Surrounding Material (Air)"
           << "\n"
           << "    "
           << " + eta = " << m_etas[0].toString() << "\n"
           << "    "
           << " + k = " << m_kappas[0].toString() << "\n";

        int layer_index = 0;
        while (layer_index < nb_layers)
        {

            ss << "    "
               << "Layer " << layer_index << "\n"
               << "    "
               << " + eta = " << m_etas[layer_index + 1].toString() << "\n"
               << "    "
               << " + k = " << m_kappas[layer_index + 1].toString() << "\n"
               << "    "
               << " + alphaU = " << m_alphaUs[layer_index] << "\n"
               << "    "
               << " + alphaV = " << m_alphaVs[layer_index] << "\n";

            layer_index++;
        }
        return ss.str();
    }

    std::string toString() const
    {

        std::ostringstream oss;
        oss << "LayeredAnisotropic[" << endl
            << formStringForLayers(nb_layers, m_etas, m_kappas, m_alphaUs, m_alphaVs) << endl
            << "]";
        return oss.str();
    }

    MTS_DECLARE_CLASS()
private:
    /* Layered structure */
    int nb_layers;
    std::vector<Spectrum> m_etas;
    std::vector<Spectrum> m_kappas;
    std::vector<Float> m_alphaUs;
    std::vector<Float> m_alphaVs;

    MicrofacetDistribution::EType m_type;
    int max_bounces;
};

MTS_IMPLEMENT_CLASS_S(LayeredAnisotropic, false, BSDF)
MTS_EXPORT_PLUGIN(LayeredAnisotropic, "Layered Anisotropic BSDF");
MTS_NAMESPACE_END
