// Mitsuba includes
#include <mitsuba/core/fresolver.h>
#include <mitsuba/hw/basicshader.h>
#include <mitsuba/render/bsdf.h>

// Local includes
#include "ior.h"
#include "microfacet.h"

// Local includes
#include "layered.hpp"

MTS_NAMESPACE_BEGIN

class LayeredAnisotropicSymmetric : public BSDF {
public:
  LayeredAnisotropicSymmetric(const Properties &props) : BSDF(props) {
    ref<FileResolver> fResolver = Thread::getThread()->getFileResolver();
    parseLayers(props, nb_layers, m_tex_etas, m_tex_kappas, m_tex_alphaUs, m_tex_alphaVs);
  }

  LayeredAnisotropicSymmetric(Stream *stream, InstanceManager *manager)
      : BSDF(stream, manager) {
    configure();
  }

  void serialize(Stream *stream, InstanceManager *manager) const {
    BSDF::serialize(stream, manager);
  }

  void configure() {
    unsigned int extraFlags = 0;

    m_components.clear();
    m_components.push_back(EGlossyReflection | EFrontSide | extraFlags);

    m_usesRayDifferentials = false;

    BSDF::configure();
  }

  /* Trace moment analysis in the layer structure for a given input angle
   */
  void computeAddingDoubling(const Vector3& wi, const Vector3& wh,
                             const Spectrum *m_etas, const Spectrum *m_kappas,
                             const Float *m_alphaUs, const Float *m_alphaVs,
                             Spectrum *coeffs,
                             Float *alphaUs, Float *alphaVs) const {

    /* Init per layer coefficients to zero */
    for(int i=0; i<nb_layers; ++i)
    {
        coeffs[i] = Spectrum(0.0f);
    }

    float cosThetaI = dot(wi, wh);
    float cosThetaT;

    /* Define the different global terms for the adding-doubling */
    Spectrum R0i(0.0f), Ri0(0.0f), T0i(1.0f), Ti0(1.0f);
    Vector2 s_r0i = {0.0f, 0.0f}, s_ri0 = {0.0f, 0.0f}, s_t0i = {0.0f, 0.0f}, s_ti0 = {0.0f, 0.0f};
    float j0i = 1.0f, ji0 = 1.0f;

    /* Iterate on all the layers and apply the adding-doubling equations */
    for(int i=0; i<nb_layers; ++i)
    {
      /* Parse layer interface's data */

      // Layer interface's roughness
      Float alphaU  = m_alphaUs[i];
      Float alphaV  = m_alphaVs[i];

      // Evaluate the potentially complex IOR.
      // In case the IOR is real, we only need a floatting point value for Snell's law.
      Spectrum eta1   = m_etas[i];
      Spectrum eta2   = m_etas[i + 1];
      Spectrum eta    = eta2 / eta1;
      Spectrum invEta = eta1 / eta2;
      float eta_f     = eta.average();
      float invEta_f  = invEta.average();
      Spectrum kappa  = m_kappas[i + 1] / eta1;


      /* Define the different local terms for the adding-doubling */
      Spectrum R12, R21, T12, T21;
      Vector2 s_r12 = {0.0f, 0.0f}, s_r21 = {0.0f, 0.0f}, s_t12 = {0.0f, 0.0f}, s_t21 = {0.0f, 0.0f};
      float j12 = 1.0f, j21 = 1.0f;


      /* Evaluate Snell' law */
      float sinThetaI = sqrt(1.0f - cosThetaI * cosThetaI);
      float sinThetaT = sinThetaI * invEta_f;
      if (sinThetaT <= 1.0f) {
        cosThetaT = sqrt(1.0f - sinThetaT * sinThetaT);
      } else {
        cosThetaT = -1.0f;
      }


      /* Ray is not block by conducting interface or total reflection */
      const bool has_transmissive = cosThetaT > 0.0f && kappa.isZero();


      /* Evaluate interface variance term */
      s_r12 = Vector2(roughnessToVariance(alphaU), roughnessToVariance(alphaV));
      s_r21 = s_r12;

      // Calculate variance transmitted from interface's roughness
      if (has_transmissive) {
        // accounting for the angular width scale of the transmitted lobe
        // affecting roughness
        s_t12 = Vector2(roughnessToVariance(alphaU * 0.5f * fabs(cosThetaT * eta_f - cosThetaI) / (cosThetaT * eta_f)),
                        roughnessToVariance(alphaV * 0.5f * fabs(cosThetaT * eta_f - cosThetaI) / (cosThetaT * eta_f)));

        s_t21 = Vector2(roughnessToVariance(alphaU * 0.5f * fabs(cosThetaI * invEta_f - cosThetaT) / (cosThetaI * invEta_f)),
                        roughnessToVariance(alphaV * 0.5f * fabs(cosThetaI * invEta_f - cosThetaT) / (cosThetaI * invEta_f)));

        // Scale due to the interface
        j12 = (cosThetaT / cosThetaI) * eta_f;
        j21 = 1.0/j12;
      }


      /* Evaluate the reflection coefficent R12 and the transmission coefficient T12 */
      R12 = (kappa.isZero()) ? fresnelDielectricExt(cosThetaI, eta_f) * Spectrum(1.0f)
                             : fresnelConductorExact(cosThetaI, eta, kappa);
      T12 = (kappa.isZero()) ? Spectrum(1.0f) - R12 : Spectrum(0.0f);

      if (has_transmissive) {
        R21 = R12;
        T21 = T12;
      } else {
        R21 = Spectrum(0.0f);
        T21 = Spectrum(0.0f);
        T12 = Spectrum(0.0f);
      }


      /* Note: we do not evaluate the TIR and FGD terms: we found that for the
       * Mitsuba implementation they did not provide any benefit.
       */


      /* Multiple scattering forms */
      const Spectrum denom  = (Spectrum(1.0f) - Ri0 * R12);
      const Float denom_avg = denom.average();
      const Spectrum m_R0i  = (denom_avg <= 0.0f) ? Spectrum(0.0f) : (T0i * R12 * Ti0) / denom;
      const Spectrum m_Ri0  = (denom_avg <= 0.0f) ? Spectrum(0.0f) : (T21 * Ri0 * T12) / denom;
      const Spectrum m_Rr   = (denom_avg <= 0.0f) ? Spectrum(0.0f) : (Ri0 * R12) / denom;

      // Evaluate the adding operator on the energy
      const Spectrum e_R0i = R0i + m_R0i;
      const Spectrum e_T0i = (T0i * T12) / denom;
      const Spectrum e_Ri0 = R21 + m_Ri0;
      const Spectrum e_Ti0 = (T21 * Ti0) / denom;

      // Scalar forms for the spectral quantities
      const float r0i   = R0i.average();
      const float e_r0i = e_R0i.average();
      const float e_ri0 = e_Ri0.average();
      const float m_r0i = m_R0i.average();
      const float m_ri0 = m_Ri0.average();
      const float m_rr  = m_Rr.average();
      const float r21   = R21.average();

      /* Evaluate the adding operator on the normalized variance */
      Vector2 _s_r0i = (r0i * s_r0i + m_r0i * (s_ti0 + j0i * (s_t0i + s_r12 + m_rr * (s_r12 + s_ri0))));
      Vector2 _s_t0i = j12 * s_t0i + s_t12 + j12 * (s_r12 + s_ri0) * m_rr;
      Vector2 _s_ri0 = (r21 * s_r21 + m_ri0 * (s_t12 + j12 * (s_t21 + s_ri0 + m_rr * (s_r12 + s_ri0))));
      Vector2 _s_ti0 = ji0 * s_t21 + s_ti0 + ji0 * (s_r12 + s_ri0) * m_rr;
      _s_r0i = (e_r0i > 0.0f) ? _s_r0i / e_r0i : Vector2(0.0f);
      _s_ri0 = (e_ri0 > 0.0f) ? _s_ri0 / e_ri0 : Vector2(0.0f);

      /* Store the coefficient and variance */
      if (m_r0i > 0.0f) {
        coeffs[i] = m_R0i;
        Vector2 sigma_i = (s_ti0 + j0i * (s_t0i + s_r12 + m_rr * (s_r12 + s_ri0)));
        alphaUs[i] = varianceToRoughness(sigma_i.x);
        alphaVs[i] = varianceToRoughness(sigma_i.y);
      } else {
        coeffs[i] = Spectrum(0.0f);
        alphaUs[i] = 0.0f;
        alphaVs[i] = 0.0f;
      }

      /* Update energy */
      R0i = e_R0i;
      T0i = e_T0i;
      Ri0 = e_Ri0;
      Ti0 = e_Ti0;

      /* Update mean */
      cosThetaI = cosThetaT;

      /* Update variance */
      s_r0i = _s_r0i;
      s_t0i = _s_t0i;
      s_ri0 = _s_ri0;
      s_ti0 = _s_ti0;

      /* Update jacobian */
      j0i *= j12;
      ji0 *= j21;

      /* Escape if a conductor is present */
      if(! has_transmissive) {
        return;
      }
    }

    return;
  }

  /* Evaluate the maps */
  inline void evalMaps(const BSDFSamplingRecord &bRec, Spectrum *etas,
                       Spectrum *kappas, Float *alphaUs, Float *alphaVs) const {

    // Fetch air values
    etas[0] = m_tex_etas[0]->eval(bRec.its);
    kappas[0] = m_tex_kappas[0]->eval(bRec.its);

    // Fetch textured values
    for (int i = 0; i < nb_layers; ++i) {
      etas[i + 1] = m_tex_etas[i + 1]->eval(bRec.its);
      kappas[i + 1] = m_tex_kappas[i + 1]->eval(bRec.its);
      alphaUs[i] = m_tex_alphaUs[i]->eval(bRec.its).average();
      alphaVs[i] = m_tex_alphaVs[i]->eval(bRec.its).average();
    }
  }

  Spectrum eval(const BSDFSamplingRecord &bRec, EMeasure measure) const {
    /* Stop if this component was not requested */
    if (measure != ESolidAngle || Frame::cosTheta(bRec.wi) <= 0 ||
        Frame::cosTheta(bRec.wo) <= 0 ||
        ((bRec.component != -1 && bRec.component != 0) ||
         !(bRec.typeMask & EGlossyReflection)))
      return Spectrum(0.0f);

    /* Evaluate the maps */
    Float m_alphaUs[nb_layers], m_alphaVs[nb_layers];
    Spectrum m_kappas[nb_layers + 1], m_etas[nb_layers + 1];
    evalMaps(bRec, m_etas, m_kappas, m_alphaUs, m_alphaVs);

    auto result =
        this->eval(bRec.wi, bRec.wo, m_etas, m_kappas, m_alphaUs, m_alphaVs);
    Assert(!result.isNaN());
    return result;
  }

  Spectrum eval(const Vector3 wi, const Vector3 wo, const Spectrum *m_etas,
                const Spectrum *m_kappas, const Float *m_alphaUs,
                const Float *m_alphaVs) const {

    /* Constants */
    const Vector3 wh = normalize(wo + wi);
    const Vector3 ft = normalize(Vector3(1.0, 0.0, 0.0) - wh.x*wh);

    /* Result */
    Spectrum f(0.0f);

    /* Evaluate the adding method to get coeffs and variances */
    Spectrum coeffs[nb_layers];
    float alphaUs[nb_layers];
    float alphaVs[nb_layers];
    computeAddingDoubling(wi, wh, m_etas, m_kappas, m_alphaUs, m_alphaVs,
                          coeffs, alphaUs, alphaVs);

    /* Sum the contribution of all the interfaces */
    for (int index = 0; index < nb_layers; ++index) {

      // Skip zero contributions
      if (coeffs[index].isZero()) {
        continue;
      }

      // Evaluate microfacet model
      MicrofacetDistribution distr(MicrofacetDistribution::EGGX, alphaUs[index],
                                   alphaVs[index], true);
      const Float D = distr.eval(wh);
      const Float G = distr.G(wi, wo, wh);

      // Add to the contribution
      f += D * G * coeffs[index] / (4.0f * Frame::cosTheta(wi));
    }

    Assert(!f.isNaN());
    return f;
  }

  Float pdf(const BSDFSamplingRecord &bRec, EMeasure measure) const {
    if (measure != ESolidAngle || Frame::cosTheta(bRec.wi) <= 0 ||
        Frame::cosTheta(bRec.wo) <= 0 ||
        ((bRec.component != -1 && bRec.component != 0) ||
         !(bRec.typeMask & EGlossyReflection)))
      return 0.0f;

    /* Calculate the reflection half-vector */
    Vector wh = normalize(bRec.wo + bRec.wi);

    /* Evaluate the maps */
    Float m_alphaUs[nb_layers], m_alphaVs[nb_layers];
    Spectrum m_kappas[nb_layers + 1], m_etas[nb_layers + 1];
    evalMaps(bRec, m_etas, m_kappas, m_alphaUs, m_alphaVs);

    /* Evaluate the adding method to get coeffs and variances */
    Spectrum coeffs[nb_layers];
    float alphaUs[nb_layers];
    float alphaVs[nb_layers];
    computeAddingDoubling(bRec.wi, Vector3(0,0,1), m_etas, m_kappas, m_alphaUs, m_alphaVs,
                          coeffs, alphaUs, alphaVs);

    /* Convert Spectral coefficients to float for pdf weighting */
    float pdf = 0.0;
    float cum_w = 0.0;
    for (int i = 0; i < nb_layers; ++i) {
      // Skip zero contributions
      if (coeffs[i].isZero()) {
        continue;
      }

      /* Evlaluate weight */
      auto weight = coeffs[i].average();
      cum_w += weight;

      /* Generate NDF */
      MicrofacetDistribution distr(MicrofacetDistribution::EGGX, alphaUs[i],
                                   alphaVs[i], true);

      /* Evaluate the pdf */
      float DG = distr.eval(wh) * distr.smithG1(bRec.wi, wh) /
                 (4.0f * Frame::cosTheta(bRec.wi));
      pdf += weight * DG;
    }

    if (cum_w > 0.0f) {
      return pdf / cum_w;
    } else {
      return 0.0f;
    }
  }

  Spectrum sample(BSDFSamplingRecord &bRec, const Point2 &sample) const {
    Float pdf;
    return this->sample(bRec, pdf, sample);
  }

  Spectrum sample(BSDFSamplingRecord &bRec, Float &pdf,
                  const Point2 &sample) const {

    if (Frame::cosTheta(bRec.wi) < 0 ||
        ((bRec.component != -1 && bRec.component != 0) ||
         !(bRec.typeMask & EGlossyReflection)))
      return Spectrum(0.0f);

    Spectrum R;

    /* Evaluate the maps */
    Float m_alphaUs[nb_layers], m_alphaVs[nb_layers];
    Spectrum m_kappas[nb_layers + 1], m_etas[nb_layers + 1];
    evalMaps(bRec, m_etas, m_kappas, m_alphaUs, m_alphaVs);

    /* Evaluate the adding method to get coeffs and variances */
    Spectrum coeffs[nb_layers];
    float alphaUs[nb_layers];
    float alphaVs[nb_layers];

    computeAddingDoubling(bRec.wi, Vector3(0,0,1), m_etas, m_kappas, m_alphaUs, m_alphaVs,
                          coeffs, alphaUs, alphaVs);

    /* Convert Spectral coefficients to floats to select BRDF lobe to sample */
    float weights[nb_layers];
    float cum_w = 0.0;
    for (int i = 0; i < nb_layers; ++i) {
      weights[i] = coeffs[i].average();
      cum_w += weights[i];
    }

    /* Select a random BRDF lobe */
    float sel_w = bRec.sampler->next1D() * cum_w - weights[0];
    int sel_i = 0;
    for (sel_i = 0; sel_w > 0.0 && sel_i < nb_layers; sel_i++) {
      sel_w -= weights[sel_i + 1];
    }

    /* Sample a microfacet normal */
    MicrofacetDistribution distr(MicrofacetDistribution::EGGX, alphaUs[sel_i],
                                 alphaVs[sel_i], true);
    Normal m = distr.sample(bRec.wi, sample, pdf);

    /* Perfect specular reflection based on the microfacet normal */
    bRec.wo = reflect(bRec.wi, m);
    if (bRec.wo.z <= 0.0f || pdf <= 0.0f) {
      return Spectrum(0.0f);
    }

    bRec.eta = 1.0f;
    bRec.sampledComponent = 0;
    bRec.sampledType = EGlossyReflection;

    /* Evaluate the MIS 'pdf' using the balance heuristic */
    pdf = 0.0;
    for (int i = 0; i < nb_layers; ++i) {
      MicrofacetDistribution distr(MicrofacetDistribution::EGGX, alphaUs[i],
                                   alphaVs[i], true);
      float DG = distr.eval(m) * distr.smithG1(bRec.wi, m) /
                 (4.0f * Frame::cosTheta(bRec.wi));
      pdf += (weights[i] / cum_w) * DG;
    }

    /* Evaluate the BRDF and the final weight */
    R = this->eval(bRec.wi, bRec.wo, m_etas, m_kappas, m_alphaUs, m_alphaVs);
    Assert(!R.isNaN());

    if (pdf > 0.0f) {
      return R / pdf;
    } else {
      return Spectrum(0.0f);
    }
  }

  void addChild(const std::string &name, ConfigurableObject *child) {
    if (child->getClass()->derivesFrom(MTS_CLASS(Texture))) {
      std::string substr = name.substr(0, 3);
      std::string substrAlpha = name.substr(0, 6);
      if (substr == "eta") {
        int eta_index = atoi(name.substr(4).c_str()) + 1;
        std::cout << name << " => " << eta_index << " / " << nb_layers
                  << std::endl;
        m_tex_etas[eta_index] = static_cast<Texture *>(child);
      } else if (substr == "kap") {
        int kappa_index = atoi(name.substr(6).c_str()) + 1;
        std::cout << name << " => " << kappa_index << " / " << nb_layers
                  << std::endl;
        m_tex_kappas[kappa_index] = static_cast<Texture *>(child);
      } else if (substrAlpha == "alphaU") {
        int alphaU_index = atoi(name.substr(7).c_str());
        std::cout << name << " => " << alphaU_index << " / " << nb_layers - 1
                  << std::endl;
        m_tex_alphaUs[alphaU_index] = static_cast<Texture *>(child);
      } else if (substrAlpha == "alphaV") {
        int alphaV_index = atoi(name.substr(7).c_str());
        std::cout << name << " => " << alphaV_index << " / " << nb_layers - 1
                  << std::endl;
        m_tex_alphaVs[alphaV_index] = static_cast<Texture *>(child);
      } else {
        // other child
      }
    } else {
      BSDF::addChild(name, child);
    }
  }

  std::string formStringForLayers() const {

    std::stringstream ss;
    ss << "    "
       << "Surrounding Material (Air)"
       << "\n"
       << "    "
       << " + eta = " << m_tex_etas[0]->getAverage().toString() << "\n"
       << "    "
       << " + k = " << m_tex_kappas[0]->getAverage().toString() << "\n";

    int layer_index = 0;
    while (layer_index < nb_layers) {

      ss << "    "
         << "Layer " << layer_index << "\n"
         << "    "
         << " + eta = " << m_tex_etas[layer_index + 1]->getAverage().toString()
         << "\n"
         << "    "
         << " + k = " << m_tex_kappas[layer_index + 1]->getAverage().toString()
         << "\n"
         << "    "
         << " + alphaU = "
         << m_tex_alphaUs[layer_index]->getAverage().toString() << "\n"
         << "    "
         << " + alphaV = "
         << m_tex_alphaVs[layer_index]->getAverage().toString() << "\n";

      layer_index++;
    }
    return ss.str();
  }

  std::string toString() const {

    std::ostringstream oss;
    oss << "LayeredAnisotropicSymmetric[" << endl
        << formStringForLayers() << endl
        << "]";
    return oss.str();
  }

  MTS_DECLARE_CLASS()
private:
  /* Layered structure */
  int nb_layers;

  /* Texture maps */
  std::vector<ref<const Texture>> m_tex_etas;
  std::vector<ref<const Texture>> m_tex_kappas;
  std::vector<ref<const Texture>> m_tex_alphaUs;
  std::vector<ref<const Texture>> m_tex_alphaVs;
};

MTS_IMPLEMENT_CLASS_S(LayeredAnisotropicSymmetric, false, BSDF)
MTS_EXPORT_PLUGIN(LayeredAnisotropicSymmetric,
                  "Symmetric Layered Anisotropic BRDF");
MTS_NAMESPACE_END
